## \example display/displaying_ensembles.py
# The script shows a couple experiments with trying to visualize an
# ensemble of structures. The ensemble is fairly tight on the assembly
# scale, but there is significant variation between the location and
# orientation of the individual proteins (which were modeled as rigid
# bodies). To save space, the models have had their sidechain atoms
# removed.

from __future__ import print_function
import IMP.display
import IMP.atom
import sys

IMP.setup_from_argv(sys.argv,
    "Experiments with trying to visualize an ensemble of structures")

Segment = IMP.algebra.Segment3D
Cylinder = IMP.algebra.Cylinder3D

# turn off internal checks to speed things up
IMP.set_check_level(IMP.USAGE)


def read(m, beyond_file):
    print("reading")
    hs = []
    for i in range(0, beyond_file):
        # create a simplified version for each chain to speed up computations
        name = IMP.display.get_example_path(
            "ensemble/aligned-" + str(i) + ".pdb")
        h = IMP.atom.read_pdb(name, m, IMP.atom.CAlphaPDBSelector())
        hr = IMP.atom.Hierarchy.setup_particle(IMP.Particle(m))
        hs.append(hr)
        for c in IMP.atom.get_by_type(h, IMP.atom.CHAIN_TYPE):
            simp = IMP.atom.create_simplified_along_backbone(
                IMP.atom.Chain(c), 4)
            hr.add_child(simp)
        IMP.atom.destroy(h)
        if i == 0:
            base = IMP.atom.get_leaves(hr)
        print(" ", i)
    return hs


def add_markers(h, c, w):
    """Add markers to a the passed conformation. The marker locations are chosen
    pretty thoughtlessly and don't really illustrate the technique well."""
    def add_marker(s, name):
        g = IMP.core.XYZRGeometry(s.get_selected_particles()[0])
        g.set_name(name)
        g.set_color(c)
        w.add_geometry(g)
    s = IMP.atom.Selection(h, chain='B', residue_index=317)
    add_marker(s, "m0")
    s = IMP.atom.Selection(h, chain='G', residue_index=212)
    add_marker(s, "m1")
    s = IMP.atom.Selection(h, chain='I', residue_index=237)
    add_marker(s, "m2")
    s = IMP.atom.Selection(h, chain='F', residue_index=101)
    add_marker(s, "m3")


def get_nice_name(h):
    nm = h.get_name()
    return nm[nm.find('-') + 1:nm.rfind('.')]


def add_axis(h, c, w, chain_colors):
    """Add a coordinate axis to show the relative orientation of the protein"""
    for hc in IMP.atom.get_by_type(h, IMP.atom.CHAIN_TYPE):
        rb = IMP.core.RigidMember(hc).get_rigid_body()
        g = IMP.display.ReferenceFrameGeometry(rb.get_reference_frame())
        g.set_name(get_nice_name(h) + "_orient")
        if c:
            g.set_color(c)
        else:
            g.set_color(chain_colors[IMP.atom.Chain(hc).get_id()])
        w.add_geometry(g)


def add_skeleton(h, c, r, w, chain_colors):
    """Show the connectivity skeleton of the conformation to give an idea of
    how things are laid out"""
    for hc0 in IMP.atom.get_by_type(h, IMP.atom.CHAIN_TYPE):
        for hc1 in IMP.atom.get_by_type(h, IMP.atom.CHAIN_TYPE):
            if hc1 <= hc0:
                continue
            d = ps.evaluate_index(h.get_model(),
                                  (hc0.get_particle_index(),
                                   hc1.get_particle_index()), None)
            if d < 1:
                d0 = IMP.core.XYZ(hc0)
                d1 = IMP.core.XYZ(hc1)
                mp = .5 * (d0.get_coordinates() + d1.get_coordinates())
                g = IMP.display.CylinderGeometry(
                    Cylinder(Segment(d0.get_coordinates(), mp), r))
                if c:
                    g.set_color(c)
                else:
                    g.set_color(chain_colors[IMP.atom.Chain(d0).get_id()])
                g.set_name(get_nice_name(h) + "_skel")
                w.add_geometry(g)
                g = IMP.display.CylinderGeometry(
                    Cylinder(Segment(d1.get_coordinates(), mp), r))
                if c:
                    g.set_color(c)
                else:
                    g.set_color(chain_colors[IMP.atom.Chain(d1).get_id()])
                g.set_name(get_nice_name(h) + "_skel")
                w.add_geometry(g)

IMP.set_log_level(IMP.TERSE)
m = IMP.Model()

# change to 46 to display all of them
hs = read(m, 3)

# used to test of two molecules are touching one another
ps = IMP.core.KClosePairsPairScore(
    IMP.core.SphereDistancePairScore(IMP.core.HarmonicUpperBound(10, 1)),
    IMP.core.LeavesRefiner(IMP.atom.Hierarchy.get_traits()))
ps.set_log_level(IMP.SILENT)


print("creating rigid bodies")
base_chains = {}
for hc in IMP.atom.get_by_type(hs[0], IMP.atom.CHAIN_TYPE):
    c = IMP.atom.Chain(hc)
    base_chains[c.get_id()] = c

for i, h in enumerate(hs):
    for hc in IMP.atom.get_by_type(h, IMP.atom.CHAIN_TYPE):
        c = IMP.atom.Chain(hc)
        if h == hs[0]:
            crb = IMP.atom.create_rigid_body(hc)
        else:
            # make sure the rigid bodies have equivalent defining reference frames
            # if we just used IMP.atom.create_rigid_body, globular proteins are likely
            # to have different axis computed when starting in different
            # orientations
            crb = IMP.atom.create_compatible_rigid_body(
                hc, base_chains[c.get_id()])
    print(" ", i)

chains = IMP.atom.get_by_type(hs[0], IMP.atom.CHAIN_TYPE)
chains.sort(key = lambda x: IMP.core.XYZ(x).get_x() + IMP.core.XYZ(x).get_y())
chain_colors = {}
for i, c in enumerate(chains):
    id = IMP.atom.Chain(c).get_id()
    #f= i/float(len(chains))
    color = IMP.display.get_display_color(i)
    # IMP.display.get_jet_color(f)
    chain_colors[id] = color

w = IMP.display.PymolWriter("markers.pym")
add_markers(hs[0], IMP.display.Color(1, 1, 1), w)
hso = hs[1:]


# sort them spatially so the colors are nicely arranged and allow one to visually connect
# the position of one end with that of the other
hso.sort(key=lambda h: IMP.core.XYZ(IMP.atom.Selection(h, chain='I',
                       residue_index=237).get_selected_particles()[0]).get_z())
print("adding markers", end=' ')
for i, h in enumerate(hso):
    c = IMP.display.get_interpolated_rgb(
        IMP.display.Color(1, 0, 0), IMP.display.Color(0, 0, 1), i / 50.)
    add_markers(h, c, w)
    print(" ", i)
w = IMP.display.PymolWriter("axis.pym")
print("adding axis", end=' ')
add_axis(hs[0], IMP.display.Color(1, 1, 1), w, chain_colors)
for i, h in enumerate(hs[1:]):
    add_axis(h, None, w, chain_colors)
    print(i, end=' ')

w = IMP.display.PymolWriter("skeletons.pym")
add_skeleton(hs[0], IMP.display.Color(1, 1, 1), 5, w, chain_colors)
print("adding skeleton", end=' ')
for i, h in enumerate(hs[1:]):
    add_skeleton(h, None, 1, w, chain_colors)
    print(" ", i)



## \example core/symmetry.py
# Show how to use the code in core to enforce symmetry.

import IMP.core
import IMP.container
import sys

IMP.setup_from_argv(sys.argv, "symmetry example")

m = IMP.Model()
m.set_log_level(IMP.SILENT)
ps = []
# create 4 xyz particles
for i in range(0, 4):
    p = IMP.Particle(m)
    ps.append(p)
    d = IMP.core.XYZ.setup_particle(p, IMP.algebra.Vector3D(.5, 0, 0))
    IMP.core.XYZR.setup_particle(p, 1)

# set the 0 particle as the reference particle for the others as
# they will get their positions from it
for i, p in enumerate(ps[1:]):
    # the other 3 particles are all symmetric copies of the first
    IMP.core.Reference.setup_particle(p, ps[0])
    # the symmetry operation is rotation around the z axis
    tr = IMP.algebra.Transformation3D(
        IMP.algebra.get_rotation_about_axis(IMP.algebra.get_basis_vector_3d(2),
                                            3.14 / 2.0 * (
                                                i + 1)),
        IMP.algebra.Vector3D(0, 0, 0))
    sm = IMP.core.TransformationSymmetry(tr)
    # set up a constraint for the one particle, if you have more than one with the same symmetry
    # transform, you should use an IMP.container.SingletonsConstraint.
    c = IMP.core.SingletonConstraint(sm, None, m, p)
    m.add_score_state(c)
lsc = IMP.container.ListSingletonContainer(m, ps)
r = IMP.core.ExcludedVolumeRestraint(lsc, 1)
sf = IMP.core.RestraintsScoringFunction([r])

d0 = IMP.core.XYZ(ps[0])
# print only optimize the main particle
d0.set_coordinates_are_optimized(True)

opt = IMP.core.ConjugateGradients(m)
opt.set_scoring_function(sf)
opt.optimize(10)
print("score is ", sf.evaluate(False))
for p in ps:
    print(p.get_name(), IMP.core.XYZ(p).get_coordinates())



#!/usr/bin/env python

from __future__ import print_function
import math
import IMP.multifit
import IMP.atom
import IMP.em
from IMP import ArgumentParser
import os
import sys

__doc__ = "Fit subunits into a density map with FFT."

multiproc_exception = None
try:
    from multiprocessing import Pool
    # Detect whether we are running Windows Python via Wine. Wine does not
    # currently support some named pipe functions which the multiprocessing
    # module needs: http://bugs.winehq.org/show_bug.cgi?id=17273
    if sys.platform == 'win32' and 'WINELOADERNOEXEC' in os.environ:
        multiproc_exception = "Wine does not currently support multiprocessing"
except ImportError as detail:
    multiproc_exception = str(detail)


class Fitter(object):

    def __init__(
        self,
        em_map,
        spacing,
        resolution,
        origin,
        density_threshold,
        pdb,
        fits_fn,
        angle,
        num_fits,
        angles_per_voxel,
            ref_pdb=''):
        self.em_map = em_map
        self.spacing = spacing
        self.resolution = resolution
        self.threshold = density_threshold
        self.originx = origin[0]
        self.originy = origin[1]
        self.originz = origin[2]
        self.pdb = pdb
        self.fits_fn = fits_fn
        self.angle = angle
        self.num_fits = num_fits
        self.angles_per_voxel = angles_per_voxel
        self.ref_pdb = ref_pdb

    def run(self):
        print("resolution is:", self.resolution)
        dmap = IMP.em.read_map(self.em_map)
        dmap.get_header().set_resolution(self.resolution)
        dmap.update_voxel_size(self.spacing)
        dmap.set_origin(IMP.algebra.Vector3D(self.originx,
                                             self.originy,
                                             self.originz))
        dmap.set_was_used(True)
        dmap.get_header().show()
        mdl = IMP.Model()
        mol2fit = IMP.atom.read_pdb(self.pdb, mdl)
        mh_xyz = IMP.core.XYZs(IMP.core.get_leaves(mol2fit))
        rb = IMP.atom.create_rigid_body(mol2fit)
        ff = IMP.multifit.FFTFitting()
        ff.set_was_used(True)
        fits = ff.do_global_fitting(dmap, self.threshold, mol2fit,
                                    self.angle / 180.0 * math.pi,
                                    self.num_fits, self.spacing, 0.5,
                                    True, self.angles_per_voxel)
        fits.set_was_used(True)
        final_fits = fits.best_fits_
        if self.ref_pdb != '':
            ref_mh = IMP.atom.read_pdb(self.ref_pdb, mdl)
            ref_mh_xyz = IMP.core.XYZs(IMP.core.get_leaves(ref_mh))
            cur_low = [1e4, 0]
        for i, fit in enumerate(final_fits):
            fit.set_index(i)
            if self.ref_pdb != '':
                trans = fit.get_fit_transformation()
                IMP.atom.transform(mol2fit, trans)
                rmsd = IMP.atom.get_rmsd(mh_xyz, ref_mh_xyz)
                if rmsd < cur_low[0]:
                    cur_low[0] = rmsd
                    cur_low[1] = i
                fit.set_rmsd_to_reference(rmsd)
                IMP.atom.transform(mol2fit, trans.get_inverse())
        if self.ref_pdb != '':
            print('from all fits, lowest rmsd to ref:', cur_low)
        IMP.multifit.write_fitting_solutions(self.fits_fn, final_fits)


def do_work(f):
    f.run()


def parse_args():
    desc = """Fit subunits into a density map with FFT."""
    p = ArgumentParser(description=desc)
    p.add_argument("-c", "--cpu", dest="cpus", type=int, default=1,
                   help="number of cpus to use (default 1)")
    p.add_argument("-a", "--angle", dest="angle", type=float, default=30,
                   help="angle delta (degrees) for FFT rotational "
                        "search (default 30)")

    p.add_argument("-n", "--num", dest="num", type=int,
                   default=100, help="Number of fits to report (default 100)")

    p.add_argument("-v", "--angle_voxel", dest="angle_voxel", type=int,
                   default=10,
                   help="Number of angles to keep per voxel (default 10)")

    p.add_argument("assembly_file", help="assembly file name")

    # p.add_argument("-n", "--num", dest="num", type="int",
    #                  default=100,
    #                  help="Number of fits to report"
    #                      "(default 100)")

    return p.parse_args()


def run(asmb_fn, options):
    if multiproc_exception is None and options.cpus > 1:
        work_units = []
    asmb_input = IMP.multifit.read_settings(asmb_fn)
    asmb_input.set_was_used(True)
    em_map = asmb_input.get_assembly_header().get_dens_fn()
    resolution = asmb_input.get_assembly_header().get_resolution()
    spacing = asmb_input.get_assembly_header().get_spacing()
    origin = asmb_input.get_assembly_header().get_origin()
    for i in range(asmb_input.get_number_of_component_headers()):
        fits_fn = asmb_input.get_component_header(i).get_transformations_fn()
        pdb_fn = asmb_input.get_component_header(i).get_filename()
        f = Fitter(
            em_map,
            spacing,
            resolution,
            origin,
            asmb_input.get_assembly_header().get_threshold(),
            pdb_fn,
            fits_fn,
            options.angle,
            options.num,
            options.angle_voxel)
        if multiproc_exception is None and options.cpus > 1:
            work_units.append(f)
        else:
            if options.cpus > 1:
                options.cpus = 1
                print("""
The Python 'multiprocessing' module (available in Python 2.6 and later) is
needed to run on multiple CPUs, and could not be found
(Python error: '%s').
Running on a single processor.""" % multiproc_exception, file=sys.stderr)
            f.run()
    if multiproc_exception is None and options.cpus > 1:
        # No point in spawning more processes than components
        nproc = min(options.cpus, asmb_input.get_number_of_component_headers())
        p = Pool(processes=nproc)
        out = list(p.imap_unordered(do_work, work_units))


def main():
    args = parse_args()
    run(args.assembly_file, args)

if __name__ == "__main__":
    main()



# Autogenerated by tools\build\make_module_version.py
# Do not edit - any changes will be lost!

def check_version(myversion):
  def _check_one(name, expected, found):
      if expected != found:
          raise RuntimeError(
              "Expected version " + expected + " but got " + found
              + " when loading module " + name
              + ". Please make sure IMP is properly built and installed "
              + "and that matching python and C++ libraries are used.")
  version = '2.15.0'
  _check_one('example', version, myversion)
  


#!/usr/bin/env python
import sys

if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    sys.stderr.write("""
Python version 2.7 or later is required to build and use IMP (any version of Python 3 can also be used).
The Python binary (found in your path at %s) is version %d.%d. Please ensure that you have a suitable version of Python available, and it is in your PATH. e.g. if you have Python 2.7 as /usr/local/bin/python27 something like the following would work:
    mkdir bin
    ln -sf /usr/local/bin/python27 bin/python
    PATH=`pwd`/bin:$PATH
    """ % (sys.executable, sys.version_info[0], sys.version_info[1]))   # noqa: E501
    sys.exit(1)



## \example core/restrain_in_sphere.py
# This fragment shows how to restrain a set of points stored in a
# SingletonContainer in a sphere of radius 'radius' centered around
# 'center'.

import IMP
import IMP.core
import sys

IMP.setup_from_argv(sys.argv, "restrain in sphere")

radius = 10
stiffness = 2
center = IMP.algebra.Vector3D(1, 2, 3)
m = IMP.Model()
p = m.add_particle("particle")
IMP.core.XYZR.setup_particle(
    m, p, IMP.algebra.Sphere3D(IMP.algebra.Vector3D(100, 0, 0), 1))

ub = IMP.core.HarmonicUpperBound(radius, stiffness)

# Restrain based on the distance to a single point (hence a ball
ss = IMP.core.DistanceToSingletonScore(ub, center)

r = IMP.core.SingletonRestraint(m, ss, p)
r.evaluate(False)



## \example parallel/tasks.py
# This module contains the setup and task functions used by local_distance.py.
#

import IMP
import IMP.algebra
import IMP.core


def setup():
    """Create a Model containing two XYZ particles linked by a harmonic
       distance restraint, one fixed at the origin. Return the Model,
       the ScoringFunction, and the free XYZ particle."""
    m = IMP.Model()
    p1 = IMP.Particle(m)
    d1 = IMP.core.XYZ.setup_particle(p1)
    p2 = IMP.Particle(m)
    d2 = IMP.core.XYZ.setup_particle(p2)
    d1.set_coordinates(IMP.algebra.Vector3D(0, 0, 0))
    r = IMP.core.DistanceRestraint(m, IMP.core.Harmonic(0, 1), p1, p2)
    sf = IMP.core.RestraintsScoringFunction([r])
    return m, sf, d2

# Note that setup and tasks are Python callables, i.e. functions (like setup
# above) or classes that implement the __call__ method (like Task below).
# The latter allows for parameters (Python objects) to be passed from the
# manager to the workers.


class Task(object):

    def __init__(self, dist):
        self.dist = dist

    def __call__(self, m, sf, d):
        """Place the free XYZ particle at the specified distance from the
           origin. Return the distance and the model's score. Note that the
           input parameters to this method (m, sf, and d) are those returned by
           the setup function above."""
        d.set_coordinates(IMP.algebra.Vector3D(0, 0, self.dist))
        return (self.dist, sf.evaluate(False))



## \example pmi/ambiguity.py
"""This script shows how to create a system with multiple copies of the same molecule.
We also create some cross-links which take into account the ambiguity.
The key to ambiguity is using the same molecule name for ambiguous copies.
That way when you perform Selection it automatically finds all relevant molecules.
"""

import IMP
import IMP.atom
import IMP.algebra
import IMP.rmf
import IMP.pmi
import IMP.pmi.topology
import IMP.pmi.dof
import IMP.pmi.macros
import IMP.pmi.restraints.stereochemistry
import IMP.pmi.restraints.crosslinking
import tempfile
import os
import sys

IMP.setup_from_argv(sys.argv, "ambiguity example")

###################### SYSTEM SETUP #####################

### Setup multistate system
mdl = IMP.Model()
s = IMP.pmi.topology.System(mdl)
st1 = s.create_state()
st2 = s.create_state()

### For each state add some molecules - we'll make some bead only structures
# State 1: ProtA (chainA), ProtA (chainB), ProtC (chainC)
sequence = 'A'*10
m1A = st1.create_molecule('ProtA',sequence,chain_id='A')
m1A.add_representation(m1A,resolutions=[1])
m1B = m1A.create_clone(chain_id='B') # create_clone() will copy name/structure/representation
                                     # You cannot edit it!
                                     # There is also a function create_copy() which
                                     #  only copies the name, then you can change reps
m1C = st1.create_molecule('ProtC',sequence,chain_id='C')
m1C.add_representation(m1C,resolutions=[1])

# State 2: ProtA (chainA), ProtC (chainC)
m2A = st2.create_molecule('ProtA',sequence,chain_id='A')
m2A.add_representation(m2A,resolutions=[1])
m2C = st2.create_molecule('ProtC',sequence,chain_id='C')
m2C.add_representation(m2C,resolutions=[1])
root_hier = s.build()

### Display all the states, molecules, representations
IMP.atom.show_with_representations(root_hier)

### Setup all molecules to move as flexible beads and super rigid bodies
#  "Super rigid bodies" aren't really rigid, it's just a mover that moves the whole body together
dof = IMP.pmi.dof.DegreesOfFreedom(mdl)
for mol in (m1A,m1B,m1C,m2A,m2C):
    dof.create_flexible_beads(mol,
                              max_trans=0.1)
    dof.create_super_rigid_body(mol)


###################### RESTRAINTS #####################
output_objects = [] # keep a list of functions that need to be reported

### Crosslinks setup
# 1) Create file. This one XL has 3 ambiguity options: State1 has 2, State2 has 1
lines = '''id,mol1,res1,mol2,res2,score
1,ProtA,3,ProtC,9,1.0
'''
tf = tempfile.NamedTemporaryFile(delete=False, mode='w')
tf.write(lines)
tf.close()

# 2) Define the columns
kw = IMP.pmi.io.crosslink.CrossLinkDataBaseKeywordsConverter()
kw.set_unique_id_key("id")
kw.set_protein1_key("mol1")
kw.set_protein2_key("mol2")
kw.set_residue1_key("res1")
kw.set_residue2_key("res2")
kw.set_id_score_key("score")
xldb = IMP.pmi.io.crosslink.CrossLinkDataBase(kw)
xldb.create_set_from_file(tf.name)
os.remove(tf.name)

# 3) Add the restraint
xlr = IMP.pmi.restraints.crosslinking.CrossLinkingMassSpectrometryRestraint(
    root_hier=root_hier,
    database=xldb,
    length=21,
    label="XL",
    resolution=1,
    slope=0.01)
xlr.add_to_model()
output_objects.append(xlr)
dof.get_nuisances_from_restraint(xlr) # needed to sample the nuisance particles (noise params)

### Connectivity keeps things connected along the backbone
crs = []
for mol in (m1A,m1B,m1C,m2A,m2C):
    cr = IMP.pmi.restraints.stereochemistry.ConnectivityRestraint(mol)
    cr.add_to_model()
    output_objects.append(cr)

### Excluded volume - one for each state (they don't interact)
evr1 = IMP.pmi.restraints.stereochemistry.ExcludedVolumeSphere(included_objects = (m1A,m1B,m1C))
evr1.add_to_model()
output_objects.append(evr1)
evr2 = IMP.pmi.restraints.stereochemistry.ExcludedVolumeSphere(included_objects = (m2A,m2C))
evr2.add_to_model()
output_objects.append(evr2)

###################### SAMPLING #####################
# randomize particles a bit
IMP.pmi.tools.shuffle_configuration(root_hier,
                                    max_translation=20)

# Shift state 2
# Even though the two states don't interact,
#  it'll be easier to look at the RMF if we separate them
trans = IMP.algebra.Transformation3D([50,0,0])
for fb in IMP.core.get_leaves(m2A.get_hierarchy())+IMP.core.get_leaves(m2C.get_hierarchy()):
    IMP.core.transform(IMP.core.XYZ(fb),trans)

# Run replica exchange Monte Carlo sampling
rex=IMP.pmi.macros.ReplicaExchange0(mdl,
                                    root_hier=root_hier,                          # pass the root hierarchy
                                    monte_carlo_sample_objects=dof.get_movers(),  # pass MC movers
                                    global_output_directory='ambiguity_output/',
                                    output_objects=output_objects,
                                    monte_carlo_steps=10,
                                    number_of_frames=1)     # increase number of frames to get better results!
rex.execute_macro()



import networkx as NX
from xml.dom import minidom, Node
#import xml.dom.ext as domExt
import sys
(NO_EDGE, EDGE_TYPE_SOLID) = range(2)


class InteractionGraph(NX.Graph):

    def init_graph(self, pairs):
        self.load_dependencies(pairs)

    def load_dependencies(self, pairs):
        for p in pairs:
            compA = p[0]
            compB = p[1]
            self.add_edge(compA, compB)



"""@namespace IMP.EMageFit.domino_model
   Classes to manage a model using DOMINO.
"""

import sys
import time
import csv
import logging
log = logging.getLogger("DominoModel")

import IMP
import IMP.domino as domino
import IMP.core as core
import IMP.container as container
import IMP.display as display
import IMP.atom as atom
import IMP.algebra as alg
import IMP.em2d as em2d
import IMP.multifit as multifit

import IMP.EMageFit.imp_general.comparisons as comparisons
import IMP.EMageFit.imp_general.alignments as alignments
import IMP.EMageFit.imp_general.io as io
import IMP.EMageFit.imp_general.representation as representation

import IMP.EMageFit.restraints as restraints
import IMP.EMageFit.buildxlinks as buildxlinks
import IMP.EMageFit.solutions_io as solutions_io
import IMP.EMageFit.csv_related as csv_related


field_delim = ","  # separate fields in the output text file
unit_delim = "/"  # separate units within a field (eg, reference frames).
                 # It is used in output text files and in databse files


class DominoModel:

    """
        Management of a model using DOMINO
    """

    def __init__(self, name="my model"):
        self.model = IMP.Model()
        self.model.set_name(name)
        self.configuration_sampling_done = False
        self.assignments_sampling_done = False
        self.rb_states_table = domino.ParticleStatesTable()
        self.merge_tree = None
        # measure_models is True only when doing benchmark
        self.measure_models = False
        # The first time that create_coarse is found and its value is True
        # a coarse version of the assembly is built
        self.create_coarse = True
        self.restraints = dict()
        self.xlinks_dict = buildxlinks.XlinksDict()

    def add_restraint(self, r, name, weight, max_score=False):
        """
            Adds a restraint to the model
            @param r An IMP.Restraint object
            @param name Name for the restraint
            @param weight Weight for the restraint
            @param max_score Maximum score allowed for the restraint. If
            max_score is False, there is no limit for the value of the restraint
        """
        log.info("Adding restraint %s weight %s max_score %s",
                 name, weight, max_score)
        if "-" in name:
            raise ValueError("SQL database does not accept restraint names "
                             "containing -")
        r.set_name(name)
        if max_score is not None and max_score:
            r.set_maximum_score(max_score)
        r.set_weight(weight)
        self.restraints[name] = r

    def align_rigid_bodies_states(self):
        """
            Aligns the set of structures considered for DOMINO sampling.
            The function:
                1) reads the possible reference frames that the
                    rb_states_table stores for each rigid body. This table
                    must be filled before using this function. Usually it is
                    filled with the results from a previous Monte Carlo sampling.
                    If the solutions from Monte Carlo seem to have the same structure
                    but they are not aligned to each other, this function can
                    help setting better starting positions to use with DOMINO.
                2) Gets the first state for each of the rigid bodies and sets
                   a reference structure using such states.
                3) Aligns all the rest of the structures respect to the reference
                4) Replaces the values of the reference frames stored in the
                   rb_states_table with the new values obtained from the alignments.
                   It does it for all states of a rigid body.
            @note: If this function is applied, the parameters "anchor" and "fixed"
               are ignored, as they are superseded by the use of the alignments
               calculated here.
        """
        log.debug("Align the configurations read from the database before "
                  "running DOMINO")
        rb_states = []
        n_states = []
        for rb in self.components_rbs:
            ps = self.rb_states_table.get_particle_states(rb)
            n = ps.get_number_of_particle_states()
            if n == 0:
                raise ValueError(
                    "There are no particle states for %s" %
                    rb.get_name())
            n_states.append(n)
            rb_states.append(ps)
        # coordinates of the first configuration (there is at least one for all
        # the rbs)
        for rb, states in zip(self.components_rbs, rb_states):
            states.load_particle_state(0, rb)
        reference_coords = get_coordinates(self.components_rbs)
        aligned_transformations = [[] for rb in self.components_rbs]
        for i in range(1, max(n_states)):
            log.debug("Aligning rigid bodies configuration %s" % i)
            # load the configuration
            for j in range(len(self.components_rbs)):
                # if there are no more particle states for this rigid body, use
                # the last
                state_index = min(i, n_states[j] - 1)
                rb_states[j].load_particle_state(
                    state_index,
                    self.components_rbs[j])
            coords = get_coordinates(self.components_rbs)
            T = alg.get_transformation_aligning_first_to_second(
                coords, reference_coords)
            for j, rb in enumerate(self.components_rbs):
                t = rb.get_reference_frame().get_transformation_to()
                new_t = alg.compose(T, t)
                aligned_transformations[j].append(new_t)
        # set the new possible transformations
        for i, rb in enumerate(self.components_rbs):
            self.set_rb_states(rb, aligned_transformations[i])

    def set_xlink_restraint(self, id1, chain1, residue1, id2, chain2, residue2,
                            distance, weight, stddev, max_score=False):
        """
            Set a restraint on the maximum distance between 2 residues
            @param id1 Name of the first component
            @param chain1
            @param residue1 Residue number for the aminoacid in the first
                component.The number is the number in the PDB file, not the
                number relative to the beginning of the chain
            @param id2 Name of the second component
            @param chain2
            @param residue2 Residue number for the aminoacid in the second
              component.
            @param distance Maximum distance tolerated
            @param weight Weight of the restraint
            @param stddev Standard deviation used to approximate the
                HarmonicUpperBound function to a Gaussian
            @param max_score See help for add_restraint(). If none is given,
            the maximum score is set to allow a maximum distance of 10 Angstrom
            greater than the parameter "distance".
        """
        xlink = buildxlinks.Xlink(
            id1,
            chain1,
            residue1,
            id2,
            chain2,
            residue2,
            distance)
        log.info("Setting cross-linking restraint ")
        log.info("%s", xlink.show())
        self.xlinks_dict.add(xlink)
        # setup restraint
        A = representation.get_component(self.assembly, xlink.first_id)
        s1 = IMP.atom.Selection(A, chain=xlink.first_chain,
                                residue_index=xlink.first_residue)
        p1 = s1.get_selected_particles()[0]
        B = representation.get_component(self.assembly, xlink.second_id)
        s2 = IMP.atom.Selection(B, chain=xlink.second_chain,
                                residue_index=xlink.second_residue)
        p2 = s2.get_selected_particles()[0]
        k = core.Harmonic.get_k_from_standard_deviation(stddev)
        score = core.HarmonicUpperBound(xlink.distance, k)
        pair_score = IMP.core.DistancePairScore(score)
        r = IMP.core.PairRestraint(self.model, pair_score, (p1, p2))
        if not max_score:
            error_distance_allowed = 100
            max_score = weight * \
                score.evaluate(distance + error_distance_allowed)
        log.info("%s, max_score %s", xlink.show(), max_score)
        self.add_restraint(r, xlink.get_name(), weight, max_score)

    def set_complementarity_restraint(self, name1, name2, rname,
                                      max_sep_distance, max_penetration,
                                      weight, max_score=1e10):
        """
            Set a restraint for geometric complementarity between 2 components
            @param name1 name of
            @param name2 - The restraint is applied to this components
            @param rname - The name given to the restraint
            @param max_sep_distance - maximum distance between molecules
                                tolerated by the restraint
            @param max_penetration - Maximum penetration allowed (angstrom)
            @param weight
            @param max_score
        """
        log.info("Setting geometric complementarity restraint %s: %s - %s",
                 rname, name1, name2)
        A = representation.get_component(self.assembly, name1)
        B = representation.get_component(self.assembly, name2)
        restraint = multifit.ComplementarityRestraint(atom.get_leaves(A),
                                                      atom.get_leaves(B), rname)
        restraint.set_maximum_separation(max_sep_distance)
        restraint.set_maximum_penetration(max_penetration)
        log.debug("Maximum separation: %s Maximum penetration score: %s",
                  max_sep_distance, max_penetration)
        self.add_restraint(restraint, rname, weight, max_score)

    def create_coarse_assembly(self, n_residues):
        """
            Simplify the assembly with a coarse representation
            @param n_residues Number of residues used for a coarse particle
        """
        if self.create_coarse:
            log.info("Creating simplified assembly")
            self.coarse_assembly = \
                representation.create_simplified_assembly(self.assembly,
                                                          self.components_rbs, n_residues)
        self.create_coarse = False

    def do_sampling(self, mode="assignments_heap_container", params=None):
        """
            Do Domino sampling
            @param mode The mode used to recover solutions from domino.
                It can have
              the value "configuration", "assignments", or
              "assignments_heap_container". The mode
              "configuration" recovers all possible configurations
              (takes a  while to generate them).
              The mode "assignments" only provides the assignments
              (state numbers) of the solutions. It is faster but requires
              generating the solutions afterwards
              The mode "assignments_heap_container" selects the best solutions
              after each merging in DOMINO, discarding the rest.
              In practice I used the mode "assignments_heap_container"
            @param params
        """
        t0 = time.time()
        if mode == "configuration":
            self.configuration_set = self.sampler.create_sample()
            tf = time.time()
            log.info("found %s configurations. Time %s",
                     self.configuration_set.get_number_of_configurations(),
                     tf - t0)
            self.configuration_sampling_done = True
        elif mode == "assignments":
            subset = self.rb_states_table.get_subset()
            log.info(
                "Do sampling with assignments. Subset has %s elements: %s",
                len(subset), subset)
            self.solution_assignments = \
                self.sampler.get_sample_assignments(subset)
            tf = time.time()
            log.info("found %s assignments. Time %s",
                     len(self.solution_assignments), tf - t0)
            self.assignments_sampling_done = True
        elif mode == "assignments_heap_container":
            subset = self.rb_states_table.get_subset()
            log.info("DOMINO sampling  an assignments_heap_container. "
                     "Subset has %s elements: %s", len(subset), subset)
            # last vertex is the root of the merge tree
            root = self.merge_tree.get_vertices()[-1]
            container = self.get_assignments_with_heap(root,
                                                       params.heap_solutions)
            self.solution_assignments = container.get_assignments()
            tf = time.time()
            log.info("found %s assignments. Time %s",
                     len(self.solution_assignments), tf - t0)
            self.assignments_sampling_done = True
        else:
            raise ValueError("Requested wrong mode for the DOMINO sampling")

    def get_assignment_text(self, assignment, subset):
        """
            Get the formatted text for an assignment
            @param subset Subset of components of the assembly
            @param assignment Assignment class with the states for the subset
            @note: The order in assignment and subset is not always the order
            of the rigid bodies in self.components_rbs. The function reorders
            the terms in the assignment so there is correspondence.
        """
        ordered_assignment = []
        for rb in self.components_rbs:
            for i, particle in enumerate(subset):
                if rb.get_particle() == particle:
                    ordered_assignment.append(assignment[i])
        text = "|".join([str(i) for i in ordered_assignment])
        return text

    def get_assignment_and_RFs(self, assignment, subset):
        """
            Return a line with texts for an assignment and the
            the reference frames of the RigidBodies in the subset.
            @param subset Subset of components of the assembly
            @param assignment Assignment class with the states for the subset
            @note: see the get_assignment_assignment_text() note.
        """
        ordered_assignment = []
        RFs = []
        for rb in self.components_rbs:
            for i, particle in enumerate(subset):
                if rb.get_particle() == particle:
                    j = assignment[i]
                    ordered_assignment.append(j)
                    pstates = self.rb_states_table.get_particle_states(rb)
                    pstates.load_particle_state(j, rb.get_particle())
                    RFs.append(rb.get_reference_frame())
        RFs_text = unit_delim.join(
            [io.ReferenceFrameToText(ref).get_text() for ref in RFs])
        assignment_numbers = "|".join([str(i) for i in ordered_assignment])
        return [assignment_numbers, RFs_text]

    def get_assignment_info(self, assignment, subset):
        """
            Info related to an assignment. Returns a list with text for the
            assignment and all the scores for the restraints
            @param subset Subset of components of the assembly
            @param assignment Assignment class with the states for the subset
        """
        text = self.get_assignment_and_RFs(assignment, subset)
        scores, total_score = self.get_assignment_scores(assignment, subset)
        info = text + [total_score] + scores
        return info

    def get_assignment_scores(self, assignment, subset):
        """
            Returns a list with the values for the restraints on a subset of
            the components of the assembly
            @param subset Subset of components of the assembly
            @param assignment Assignment class with the states for the subset
        """
        restraints_to_evaluate = \
            self.restraint_cache.get_restraints(subset, [])
        scores = [self.restraint_cache.get_score(r, subset, assignment)
                  for r in restraints_to_evaluate]
        total_score = sum(scores)
        return scores, total_score

    def load_state(self, assignment):
        """
            Load the positions of the components given by the assignment
            @param assignment Assignment class
        """
        subset = self.rb_states_table.get_subset()
        domino.load_particle_states(subset, assignment, self.rb_states_table)

    def get_assignments_with_heap(self, vertex, k=0):
        """
            Domino sampling that recovers the assignments for the root of the
            merge tree, but
            conserving only the best k scores for each vertex of the tree.
            @param[in] vertex Vertex with the root of the current merge tree. This
            function is recursive.
            @param[in] k
        """
        if(self.sampler.get_number_of_subset_filter_tables() == 0):
            raise ValueError("No subset filter tables")
        if(self.merge_tree is None):
            raise ValueError("No merge tree")
        # In the merge tree, the names of the vertices are the subsets.
        # The type of the vertex name is a domino.Subset
        subset = self.merge_tree.get_vertex_name(vertex)
        log.info("computing assignments for vertex %s", subset)
        t0 = time.time()
        assignments_container = domino.HeapAssignmentContainer(subset, k,
                                                               self.restraint_cache, "my_heap_assignments_container")
        neighbors = self.merge_tree.get_out_neighbors(vertex)
        if len(neighbors) == 0:  # A leaf
            # Fill the container with the assignments for the leaf
            self.sampler.load_vertex_assignments(vertex, assignments_container)
        else:
            if neighbors[0] > neighbors[1]:
                # get_vertex_assignments() methods in domino
                # expects the children in sorted order
                neighbors = [neighbors[1], neighbors[0]]
            # recurse on the two children
            container_child0 = self.get_assignments_with_heap(neighbors[0], k)
            container_child1 = self.get_assignments_with_heap(neighbors[1], k)
            self.sampler.load_vertex_assignments(vertex,
                                                 container_child0,
                                                 container_child1,
                                                 assignments_container)
        tf = time.time() - t0
        log.info("Merge tree vertex: %s assignments: %s Time %s sec.", subset,
                 assignments_container.get_number_of_assignments(), tf)
        return assignments_container

    def get_restraint_value_for_assignment(self, assignment, name):
        """
            Recover the value of a restraint
            @param name of the restraint
            @param assignment Assignment class containing the assignment for
                        the solution desired
        """
        if not self.assignments_sampling_done:
            raise ValueError("DominoModel: sampling not done")
        log.debug("Computing restraint value for assignment %s", assignment)
        self.load_state(assignment)
        for rb in self.components_rbs:
            log.debug("rb - %s",
                      rb.get_reference_frame().get_transformation_to())
        val = self.restraints[name].evaluate(False)
        log.debug("restraint %s = %s", restraint.get_name(), val)
        return val

    def set_native_assembly_for_benchmark(self, params):
        """
            Sets the native model for benchmark, by reading the native
            structure and set the rigid bodies.
        """
        self.measure_models = True
        self.native_model = IMP.Model()
        if hasattr(params.benchmark, "fn_pdb_native"):
            self.native_assembly = \
                representation.create_assembly_from_pdb(self.native_model,
                                                        params.benchmark.fn_pdb_native, params.names)
        elif hasattr(params.benchmark, "fn_pdbs_native"):
            self.native_assembly = \
                representation.create_assembly(self.native_model,
                                               params.benchmark.fn_pdbs_native, params.names)
        else:
            raise ValueError("set_native_assembly_for_benchmark: Requested "
                             "to set a native assembly for benchmark but did not provide "
                             "its pdb, or the pdbs of the components")
        self.native_rbs = representation.create_rigid_bodies(
            self.native_assembly)
        anchor_assembly(self.native_rbs, params.anchor)

    def set_rb_states(self, rb, transformations):
        """
            Add a set of reference frames as possible states for a rigid body
            @param rb The rigid body
            @param transformations Transformations are used to build the
                                      reference frames for the rigid body.
        """
        log.info("Set rigid body states for %s", rb.get_name())
        RFs = [alg.ReferenceFrame3D(T) for T in transformations]
        for st in RFs:
            log.debug("%s", st)
        rb_states = domino.RigidBodyStates(RFs)
        self.rb_states_table.set_particle_states(rb, rb_states)
        st = self.rb_states_table.get_particle_states(rb)
        log.info("RigidBodyStates created %s",
                 st.get_number_of_particle_states())

    def add_restraint_score_filter_table(self):
        """
            Add a RestraintScoreSubsetFilterTable to DOMINO that allows to
            reject assignments that have score worse than the thresholds for
            the restraints
        """
        log.info("Adding RestraintScoreSubsetFilterTable")
        # Restraint Cache is a centralized container of the restraints and
        # their scores
        self.restraint_cache = domino.RestraintCache(self.rb_states_table)
        self.restraint_cache.add_restraints(list(self.restraints.values()))
        rssft = domino.RestraintScoreSubsetFilterTable(self.restraint_cache)
        rssft.set_name('myRestraintScoreSubsetFilterTable')
        self.sampler.add_subset_filter_table(rssft)

    def set_assembly_components(self, fn_pdbs, names):
        """
            Sets the components of an assembly, each one given as a separate
            PDB file.
            @param fn_pdbs List with the names of the pdb files
            @param names List with the names of the components of the assembly.
        """
        if len(fn_pdbs) != len(names):
            raise ValueError("Each PDB file needs a name")
        self.names = names
        self.assembly = representation.create_assembly(self.model, fn_pdbs,
                                                       names)
        self.components_rbs = representation.create_rigid_bodies(self.assembly)
        self.not_optimized_rbs = []

    def set_assembly(self, fn_pdb, names):
        """
            Sets an assembly from a single PDB file. The function assumes that
            the components of the assembly are the chains of the PDB file.
            @param fn_pdb PDB with the file for the assembly
            @param names Names for each of the components (chains)
                of the assembly
        """
        self.assembly = representation.create_assembly_from_pdb(self.model,
                                                                fn_pdb,
                                                                names)
        self.components_rbs = representation.create_rigid_bodies(self.assembly)
        self.not_optimized_rbs = []

    def add_branch_and_bound_assignments_table(self):
        """
            BranchAndBoundAssignmentsTable enumerate states based on provided
            filtered using the provided the subset filter tables
        """
        log.info("Setting BranchAndBoundAssignmentsTable")
        fts = []
        for i in range(self.sampler.get_number_of_subset_filter_tables()):
            ft = self.sampler.get_subset_filter_table(i)
            fts.append(ft)
        at = domino.BranchAndBoundAssignmentsTable(self.rb_states_table, fts)
        self.sampler.set_assignments_table(at)
        self.assignments_table = at

    def add_exclusion_filter_table(self):
        """
           ExclusionSubsetFilterTable ensures that two particles are not given
           the same state at the same time
        """
        log.info("Setting ExclusionSubsetFilterTable")
        ft = domino.ExclusionSubsetFilterTable(self.rb_states_table)
        self.sampler.add_subset_filter_table(ft)

    def setup_domino_sampler(self):
        """
            Creates a DOMINO sampler and adds the required filter tables
        """
        if self.merge_tree is None:
            raise ValueError("Merge tree for DOMINO not set. It is required "
                             "to setup the sampler")
        log.info("Domino sampler")
        self.sampler = domino.DominoSampler(self.model, self.rb_states_table)
        self.sampler.set_restraints(list(self.restraints.values()))
        self.sampler.set_log_level(IMP.TERSE)
        self.sampler.set_merge_tree(self.merge_tree)
        self.add_exclusion_filter_table()
        self.add_restraint_score_filter_table()
        self.add_branch_and_bound_assignments_table()

    def read_merge_tree(self, fn):
        """
            Read and create a merge tree from a file.
            The format of the file is the format written by write merge_tree()
            It does not matter the order of the indices in the file, as
            they are sorted by this function.
            The only condition is that the index for the vertex that is the
            root of the tree MUST be the largest. This is guaranteed when
            creating a merge tree with create_merge_tree()
            @param fn File containing the merge tree
        """
        log.info("Reading merge tree from %s", fn)
        ps = self.rb_states_table.get_particles()
        log.debug("particles")
        for p in ps:
            log.debug("%s", p.get_name())
        rows = csv_related.read_csv(fn, delimiter=field_delim)
        for i in range(len(rows)):
            row = rows[i]
            rows[i] = [int(row[0]), int(row[1]), int(row[2]), row[3]]
        # Sort rows by vertice index
        rows.sort()
        subsets = []
        # build subsets from the rows text
        for row in rows:
            names = row[3].split(unit_delim)
            log.debug("row %s names %s", row, names)
            # get particles in the subset
            particles = []
            for name in names:
                l = [p for p in ps if p.get_name() == name]
                if (len(l) > 1):
                    ValueError("More than one particle with same name" % names)
                particles.extend(l)
            s = domino.Subset(particles)
            subset_names = [p.get_name() for p in particles]
            log.debug("Merge tree Subset %s. Particles %s ", s, subset_names)
            subsets.append(s)
        # The vertex with the largest index is the root.
        # trick: get the merge tree object from a tree with only one node ...
        jt = domino.SubsetGraph()
        jt.add_vertex(subsets[0])
        mt = domino.get_merge_tree(jt)
        # ... and add the rest of the vertices
        for i in range(1, len(subsets)):
            mt.add_vertex(subsets[i])  # the name of the vertex is a subset
        # set edges
        for row in rows:
            vid = row[0]
            child_left = row[1]
            child_right = row[2]
            if child_left != -1:
                mt.add_edge(vid, child_left)
            if child_right != -1:
                mt.add_edge(vid, child_right)
        self.merge_tree = mt
        log.info("%s", self.merge_tree.show_graphviz())

    def create_merge_tree(self):
        """
            Creates a merge tree from the restraints that are set already
        """
        rs = list(self.restraints.values())
        ig = domino.get_interaction_graph(rs, self.rb_states_table)
#        pruned_dep = IMP.get_pruned_dependency_graph(self.model)
#        IMP.show_graphviz(pruned_dep)
#        IMP.show_graphviz(ig)
        jt = domino.get_junction_tree(ig)
#        IMP.show_graphviz(jt)
        self.merge_tree = domino.get_balanced_merge_tree(jt)
#        IMP.show_graphviz(self.merge_tree)
        log.info("Balanced merge tree created")
        log.info("%s", self.merge_tree.show_graphviz())

    def write_merge_tree(self, fn):
        """
            Writes the current merge tree to file. The format is:
            parent | child_left | child_right | labels
            @param fn File for storing the merge tree
        """
        log.info("Writing merge to %s", fn)
        if self.merge_tree is None:
            raise ValueError("No merge tree")
        f = open(fn, "w")
        f.write(
            "# Vertex index | Child0 | Child1 | label (names of the particles)\n")
        w = csv.writer(f, delimiter=field_delim)
        root = self.merge_tree.get_vertices()[-1]
        self.write_subset(root, w)
        f.close()

    def write_subset(self, v, w):
        """
            Writes lines describing a subset (vertex of a merge tree)
            @param w A csv.writer
            @param v Vertex index
        """
        no_child_index = -1
        subset = self.merge_tree.get_vertex_name(v)
        names = [p.get_name() for p in subset]
        reps = [x for x in names if names.count(x) > 1]  # repeated names
        if(len(reps)):
            raise ValueError("The names of the "
                             "particles in the subset %s are not unique" % subset)
        names = unit_delim.join(names)
        neighbors = self.merge_tree.get_out_neighbors(v)
        if len(neighbors) == 0:  # A leaf
            # Fill the container with the assignments for the leaf
            w.writerow([v, no_child_index, no_child_index, names])
        else:
            if neighbors[0] > neighbors[1]:
                neighbors = [neighbors[1], neighbors[0]]
            w.writerow([v, neighbors[0], neighbors[1], names])
            self.write_subset(neighbors[0], w)
            self.write_subset(neighbors[1], w)

    def set_connectivity_restraint(self, names, rname,
                                   distance=10,
                                   weight=1.0, max_score=None, n_pairs=1,
                                   stddev=2):
        """
            Set a connectivity restraint between the elements of the assembly
            @param names List with all the elements to be connected
            @param distance  Maximum distance tolerated by the restraint
            @param rname Name for the restraint
            @param weight Weight for the restraint
            @param max_score Maximum score for the restraint
            @param n_pairs Number of pairs of particles used in the
                        KClosePairScore of the connectivity restraint
            @param stddev Standard deviation used to approximate the
                                HarmonicUpperBound function to a Gaussian
        """
        components = []
        for name in names:
            c = representation.get_component(self.coarse_assembly, name)
            components.append(c)
        log.info("Connectivity restraint for %s", components)
        spring_constant = core.Harmonic.get_k_from_standard_deviation(stddev)
        if max_score is None:
            max_score = weight * spring_constant * n_pairs * (stddev) ** 2
        r = restraints.get_connectivity_restraint(
            components, distance, n_pairs,
            spring_constant)
        self.add_restraint(r, rname, weight, max_score)

    def set_em2d_restraint(self, name, fn_images_sel, pixel_size, resolution,
                           n_projections, weight, max_score=False,
                           mode="fast", n_optimized=1):
        """
            Set a Restraint computing the em2d score.
            @param name Name for the restraint
            @param fn_images_sel Selection file with the names of the images
                            used for the restraint
            @param pixel_size - pixel size in the images
            @param resolution - resolution used to downsample the projections
                    of the model when projecting
            @param weight Weight of the restraint
            @param max_score - Maximum value tolerated for the restraint
            @param n_projections - Number of projections to generate
                    when projecting the model to do the coarse alignments
            @param mode - Mode used for computing the restraints.
            @param n_optimized - number of results from the coarse
                alignment that are refined with the Simplex algorithm
                to get a more accurate value for the restraint
        """
        log.info("Adding a em2D restraint: %s", fn_images_sel)
        restraint_params = em2d.Em2DRestraintParameters(pixel_size,
                                                        resolution,
                                                        n_projections)
        r = restraints.get_em2d_restraint(self.assembly,
                                          fn_images_sel,
                                          restraint_params,
                                          mode, n_optimized)
        self.add_restraint(r, name, weight, max_score)

    def set_not_optimized(self, name):
        """
            Set a part of the model as not optimized (it does not move during
            the model optimization)
            @param name of the component to optimized
        """
        if name not in self.names:
            raise ValueError("DominoModel: There is not component "
                             "in the assembly with this name")
        rb_name = representation.get_rb_name(name)
        self.not_optimized_rbs.append(rb_name)

    def set_close_pairs_excluded_volume_restraint(self, distance=10,
                                                  weight=1.0, ratio=0.05, stddev=2,
                                                  max_score=False):
        """
            Creates an excluded volume restraint between all the components
            of the coarse assembly.See help for
            @param distance Maximum distance tolerated between particles
            @param weight Weight for the restraint
            @param stddev
            @param max_score Maximum value for the restraint. If the parameter
                is None, an automatic value is computed (using the ratio).
            @param ratio Fraction of the number of possible pairs of
                  particles that are tolerated to overlap. The maximum
                  score is modified according to this ratio.
                  I have found that ratios of 0.05-0.1 work well allowing for
                  some interpenetration
            @param stddev Standard deviation used to approximate the
                   HarmonicUpperBound function to a Gaussian
        """
        k = core.Harmonic.get_k_from_standard_deviation(stddev)
        for i, c1 in enumerate(self.coarse_assembly.get_children()):
            for j, c2 in enumerate(self.coarse_assembly.get_children()):
                if j > i:
                    name = "exc_vol_%s_%s" % (c1.get_name(), c2.get_name())

                    ls1 = atom.get_leaves(c1)
                    ls2 = atom.get_leaves(c2)
                    possible_pairs = len(ls1) * len(ls2)
                    n_pairs = possible_pairs * ratio

                    marker1 = IMP.Particle(
                        self.model, "marker1 " + name)
                    marker2 = IMP.Particle(
                        self.model, "marker2 " + name)
                    table_refiner = core.TableRefiner()
                    table_refiner.add_particle(marker1, ls1)
                    table_refiner.add_particle(marker2, ls2)
                    score = core.HarmonicLowerBound(distance, k)
                    pair_score = core.SphereDistancePairScore(score)
                    close_pair_score = core.ClosePairsPairScore(pair_score,
                                                                table_refiner,
                                                                distance)
                    r = core.PairRestraint(self.model, close_pair_score,
                                           (marker1, marker2))

                    if not max_score:
                        minimum_distance_allowed = 0
                        max_score = weight * n_pairs * \
                            score.evaluate(minimum_distance_allowed)
                    log.debug("Setting close_pairs_excluded_volume_restraint(), "
                              "max_score %s", max_score)
                    self.add_restraint(r, name, weight, max_score)

    def set_pair_score_restraint(self, name1, name2,
                                 restraint_name, distance=10,
                                 weight=1.0, n_pairs=1, stddev=2, max_score=None):
        """
            Set a pair_score restraint between the coarse representation
            of two components
            @param name1 Name of the first component
            @param name2 Name of the second component
            @param restraint_name Name for the restraint
            @param distance Maximum distance tolerated between particles
            @param weight Weight of the restraint
            @param n_pairs
            @param max_score Maximum value tolerated for the restraint
            @param stddev Standard deviation used to approximate the
                HarmonicUpperBound function to a Gaussian
        """
        table_refiner = core.TableRefiner()

        # The particles in A are attached to a marker particle via a refiner.
        # When the refiner gets a request for marker1, it returns the attached
        # particles
        A = representation.get_component(self.coarse_assembly, name1)
        marker1 = IMP.Particle(self.model, "marker1 " + restraint_name)
        table_refiner.add_particle(marker1, atom.get_leaves(A))
        # same for B
        B = representation.get_component(self.coarse_assembly, name2)
        marker2 = IMP.Particle(self.model, "marker2 " + restraint_name)
        table_refiner.add_particle(marker2, atom.get_leaves(B))

        k = core.Harmonic.get_k_from_standard_deviation(stddev)
        score = core.HarmonicUpperBoundSphereDistancePairScore(distance, k)
        # The score is set for the n_pairs closest pairs of particles
        pair_score = core.KClosePairsPairScore(score, table_refiner, n_pairs)
        #  When KClosePairsPairScore is called, the refiner will provide the
        # particles for A and B
        if not max_score:
            # Build a maximum score based on the function type that is used,
            # an HarmonicUpperBound
            temp_score = core.HarmonicUpperBound(distance, k)
            error_distance_allowed = 10
            max_score = weight * n_pairs * \
                temp_score.evaluate(distance + error_distance_allowed)

        log.info("Setting pair score restraint for %s %s. k = %s, max_score "
                 "= %s, stddev %s", name1, name2, k, max_score, stddev)
        r = core.PairRestraint(self.model, pair_score, (marker1, marker2))
        self.add_restraint(r, restraint_name, weight, max_score)

    def write_solutions_database(self, fn_database, max_number=None):
        """
            Write the results of the DOMINO sampling to a SQLite database.
            @param fn_database
            @param max_number Maximum number of results to write
        """
        log.info("Creating the database of solutions")
        if self.measure_models and not hasattr(self, "native_model"):
            raise ValueError("The native model has not been set")
        db = solutions_io.ResultsDB()
        db.create(fn_database, overwrite=True)
        db.connect(fn_database)
        subset = self.rb_states_table.get_subset()
        restraints_names = self.get_restraints_names_used(subset)
        db.add_results_table(restraints_names, self.measure_models)
        n = len(self.solution_assignments)
        if max_number is not None:
            n = min(n, max_number)
        for sol_id, assignment in enumerate(self.solution_assignments[0:n]):
            txt = self.get_assignment_text(assignment, subset)
            # load the positions of the components in the state
            self.load_state(assignment)
            RFs = [rb.get_reference_frame() for rb in self.components_rbs]
            scores, total_score = self.get_assignment_scores(
                assignment, subset)
            measures = None
            if(self.measure_models):
                measures = measure_model(self.assembly, self.native_assembly,
                                         self.components_rbs, self.native_rbs)
            db.add_record(sol_id, txt, RFs, total_score, scores, measures)
        # Add native state if doing benchmarking
        if self.measure_models:
            assignment = "native"
            RFs = [rb.get_reference_frame() for rb in self.native_rbs]
            scores, total_score = self.get_restraints_for_native(
                restraints_names)
            db.add_native_record(assignment, RFs, total_score, scores)
        db.save_records()
        db.close()

    def get_restraints_names_used(self, subset):
        """ Get the names of the restraints used for a subset
            @param subset Subset of components of the assembly
        """
        return [r.get_name() for r
                in self.restraint_cache.get_restraints(subset, [])]

    def get_restraints_for_native(self, restraints_names):
        """
            Get values of the restraints for the native structure
            @param restraints_names Names of the restraints requested
            @return a list with the values of all scores, and the total score
        """
        saved = [rb.get_reference_frame() for rb in self.components_rbs]
        # Set the state of the model to native
        for rb, rn in zip(self.components_rbs, self.native_rbs):
            # remove sub-rigid bodies
            rb_members = [m for m in rb.get_members()
                          if not core.RigidBody.get_is_setup(m.get_particle())]
            rn_members = [m for m in rn.get_members()
                          if not core.RigidBody.get_is_setup(m.get_particle())]
            rb_coords = [m.get_coordinates() for m in rb_members]
            rn_coords = [m.get_coordinates() for m in rn_members]

            # align and put rb in the position of rn
            if len(rn_coords) != len(rb_coords):
                raise ValueError("Mismatch in the number of members. "
                                 "Reference %d Aligned %d " % (len(rn_coords), len(rb_coords)))
            T = alg.get_transformation_aligning_first_to_second(rb_coords,
                                                                rn_coords)
            t = rb.get_reference_frame().get_transformation_to()
            new_t = alg.compose(T, t)
            rb.set_reference_frame(alg.ReferenceFrame3D(new_t))
        scores = []
        for name in restraints_names:
            scores.append(self.restraints[name].evaluate(False))
        total_score = sum(scores)
        representation.set_reference_frames(self.components_rbs, saved)
        return scores, total_score

    def write_pdb_for_assignment(self, assignment, fn_pdb):
        """
            Write the solution in the assignment
            @param assignment Assignment class with the states for the subset
            @param fn_pdb File to write the model
        """
        if not self.assignments_sampling_done:
            raise ValueError("DominoModel: sampling not done")
        log.info("Writting PDB %s for assignment %s", fn_pdb, assignment)
        self.load_state(assignment)
        atom.write_pdb(self.assembly, fn_pdb)

    def write_pdb_for_configuration(self, n, fn_pdb):
        """
            Write a file with a configuration for the model (configuration
            here means a configuration in DOMINO)
            @param n Index of the configuration desired.
            @param fn_pdb
        """
        if not self.configuration_sampling_done:
            raise ValueError("DominoModel: sampling not done")
        log.debug("Writting PDB %s for configuration %s", fn_pdb, n)
        configuration_set.load_configuration(n)
        atom.write_pdb(self.assembly, fn_pdb)

    def write_pdb_for_reference_frames(self, RFs, fn_pdb):
        """
            Write a PDB file with a solution given by a set of reference frames
            @param RFs Reference frames for the elements of the complex
            @param fn_pdb File to write the solution
        """
        log.debug("Writting PDB %s for reference frames %s", fn_pdb, RFs)
        representation.set_reference_frames(self.components_rbs, RFs)
        atom.write_pdb(self.assembly, fn_pdb)

    def write_pdbs_for_reference_frames(self, RFs, fn_base):
        """
            Write a separate PDB for each of the elements
            @param RFs Reference frames for the elements of the complex
            @param fn_base base string to build the names of the PDBs files
        """
        log.debug("Writting PDBs with basename %s", fn_base)
        representation.set_reference_frames(self.components_rbs, RFs)
        for i, ch in enumerate(self.assembly.get_children()):
            atom.write_pdb(ch, fn_base + "component-%02d.pdb" % i)

    def write_pdb_for_component(self, component_index, ref, fn_pdb):
        """
            Write one component of the assembly
            @param component_index Position of the component in the assembly
            @param ref Reference frame of the component
            @param fn_pdb Output PDB
        """
        ("Writting PDB %s for component %s", fn_pdb, component_index)
        self.components_rbs[component_index].set_reference_frame(ref)
        hierarchy_component = self.assembly.get_child(component_index)
        atom.write_pdb(hierarchy_component, fn_pdb)

    def get_scoring_function(self):
        """Get a ScoringFunction that includes all restraints."""
        return IMP.core.RestraintsScoringFunction(
                                  list(self.restraints.values()))

    def write_monte_carlo_solution(self, fn_database):
        """
            Write the solution of a MonteCarlo run
            @param fn_database Database file with the solution.
                                The database will contain only one record
        """
        total_score = 0
        rnames = []
        scores = []
        for r in self.restraints.values():
            score = r.evaluate(False)
            rnames.append(r.get_name())
            scores.append(score)
            total_score += score
        db = solutions_io.ResultsDB()
        db.create(fn_database, overwrite=True)
        db.connect(fn_database)
        db.add_results_table(rnames, self.measure_models)
        RFs = [rb.get_reference_frame() for rb in self.components_rbs]
        measures = None
        if(self.measure_models):
            measures = measure_model(self.assembly, self.native_assembly,
                                     self.components_rbs, self.native_rbs)
        solution_id = 0
        db.add_record(solution_id, "", RFs, total_score, scores, measures)
        db.save_records()
        db.close()

    def print_restraints_values(self):
        print("Restraints: Name, weight, value, maximum_value")
        total_score = 0
        for r in self.restraints.values():
            score = r.evaluate(False)
    #       print "%20s %18f %18f %18f" % (r.get_name(), r.get_weight(),
    #                                      score, r.get_maximum_score())
            print("%20s %18f %18f" % (r.get_name(), r.get_weight(), score))
            total_score += score
        print("total_score:", total_score)


def anchor_assembly(components_rbs, anchored):
    """
        "Anchor" a set of rigid bodies, by setting the position of one of them
        at the origin (0,0,0).
        @param components_rbs Rigid bodies of the components
        @param anchored - List of True/False values indicating if the components
      of the assembly are anchored. The function sets the FIRST anchored
      component in the (0,0,0) coordinate and moves the other
      components with the same translation. If all the values are False, the
      function does not do anything
    """
    if True in anchored:
        anchored_rb = None
        for a, rb in zip(anchored, components_rbs):
            if a:
                anchored_rb = rb
                break
        center = anchored_rb.get_coordinates()
        log.info("Anchoring assembly at (0,0,0)")
        for rb in components_rbs:
            T = rb.get_reference_frame().get_transformation_to()
            t = T.get_translation()
            R = T.get_rotation()
            log.debug("%s: Rerefence frame BEFORE %s",
                      rb.get_name(), rb.get_reference_frame())
            T2 = alg.Transformation3D(R, t - center)
            rb.set_reference_frame(alg.ReferenceFrame3D(T2))
            log.debug("%s Rerefence frame AFTER %s",
                      rb.get_name(), rb.get_reference_frame())


def measure_model(assembly, native_assembly, rbs, native_rbs):
    """
        Get the drms, cdrms and crmsd for a model
        @param assembly Hierarchy for an assembly
        @param native_assembly Hierarchy of the native structure
        @param rbs Rigid bodies for the components of the assembly
        @param native_rbs Rigid bodies for the components of the native
                    structure
        @return cdrms, cdrms, crmsd
        - drms - typical drms measure
        - cdrms - drms for the centroids of the rigid bodies of the components
          of the assembly
        - crmsd - rmsd for the centroids
    """
    log.debug("Measure model")
    drms = comparisons.get_drms_for_backbone(assembly, native_assembly)
    RFs = [rb.get_reference_frame() for rb in rbs]
    vs = [r.get_transformation_to().get_translation() for r in RFs]
    nRFs = [rb.get_reference_frame() for rb in native_rbs]
    nvs = [r.get_transformation_to().get_translation() for r in nRFs]
    cdrms = atom.get_drms(vs, nvs)
    crmsd, RFs = alignments.align_centroids_using_pca(RFs, nRFs)
    log.debug("drms %s cdrms %s crmsd %s", drms, cdrms, crmsd)
    return [drms, cdrms, crmsd]


def get_coordinates(rigid_bodies):
    """
        Return a list of the coordinates of all the members of the rigid bodies
    """
    if len(rigid_bodies) == 0:
        raise ValueError(
            "get_coordinates: There are not rigid bodies to get coordinates from")
    coords = []
    for rb in rigid_bodies:
            # remove sub-rigid bodies
        rb_members = [m for m in rb.get_members()
                      if not core.RigidBody.get_is_setup(m.get_particle())]
        coords.extend([m.get_coordinates() for m in rb_members])
    return coords



"""@namespace IMP.EMageFit.monte_carlo
   Classes for Monte Carlo sampling using rigid body relative moves.
"""

import IMP
import IMP.core as core
import IMP.atom as atom
import IMP.em2d as em2d
import IMP.em as em
import IMP.algebra as alg
import IMP.container as container
import IMP.EMageFit.imp_general.representation as representation
import IMP.EMageFit.imp_general.io as io
import sys
import os
import time
import logging
log = logging.getLogger("MonteCarloParticleSates")


#

# Generate possible configurations for the rigid bodies of the
# elements of a complex applying MonteCarlo

#


def set_random_seed(index):
    """
        initializing the random number generator manually is useful for the
        jobs in the cluster.
        @param index Seed for the random number generator
    """
    if(index != -1):
        log.debug("Seed for the random number generator: %s", index)
        IMP.random_number_generator.seed(index)
    else:
        IMP.random_number_generator.seed(time.time())


class MonteCarloRelativeMoves:

    """
        Class to do Monte Carlo sampling by using as the set of movements
        relative positions between rigid bodies
    """

    def __init__(self, model, rigid_bodies, anchored):
        log.info("Setting MonteCarloRelativeMoves")
        self.model = model
        self.rbs = rigid_bodies
        self.components = []
        self.best_models = []
        self.anchored = anchored
        self.parent_rbs = []
        # triplets with the information to build a relative mover using the
        # results from docking with HEX
        self.dock_transforms = None
        self.non_relative_move_prob = 0.1

        log.debug("Anchored components %s", self.anchored)
        T = alg.Transformation3D(alg.get_identity_rotation_3d(),
                                 alg.Vector3D(0., 0., 0.))
        origin = alg.ReferenceFrame3D(T)
        for rb in self.rbs:
            rb.set_reference_frame(origin)

    def set_temperature_pattern(self, temperatures, iterations, cycles):
        """
            Set the temperature pattern to use during the Monte Carlo
            optimizations.
            @param temperatures  List of temperature values
            @param iterations List with the iterations to do for each of the
                temperatures.
            @param cycles Number of repetitions of the
                pattern of temperatures and iterations
        """
        self.temperatures = temperatures
        self.iterations = iterations
        self.cycles = cycles
        log.debug("Temperatures %s", self.temperatures)
        log.debug("Iterations %s", self.iterations)
        log.debug("Cycles %s", self.cycles)

    def set_movers(self, max_translation, max_rotation):
        log.info("Setting movers for subunits")
        self.movers = []
        for is_anchored, rb in zip(self.anchored, self.rbs):
            if(not is_anchored):
                self.movers.append(core.RigidBodyMover(rb, max_translation,
                                                       max_rotation))

    def set_relative_movers(self, max_translation, max_rotation):
        """
            Generate the relative models form the transforms. The transforms
            is a list with triplets [id1, id2, transform_file]
            @param max_translation Maximum translation distance allowed for
                the moves
            @param max_rotation Maximum rotation angle allowed for the moves
        """
        log.info("Setting relative movers")
        self.movers = []
        relative_movers = []
        relative_names = []
        for d in self.dock_transforms:
            rb_id = representation.get_rb_name(d[1])
            if rb_id not in relative_names:
                log.debug("Checking for %s", rb_id)
                rb_lig = representation.get_rigid_body(self.rbs, rb_id)
                rb_lig.set_coordinates_are_optimized(True)
                mv = em2d.RelativePositionMover(rb_lig, max_translation,
                                                max_rotation)
                relative_movers.append(mv)
                relative_names.append(rb_id)
                log.debug("added a RelativePositionMover for %s", rb_id)
            i = relative_names.index(rb_id)
            relative_movers[i].set_random_move_probability(
                self.non_relative_move_prob)
            rb_rec = representation.get_rigid_body(self.rbs,
                                                   representation.get_rb_name(d[0]))
            rb_rec.set_coordinates_are_optimized(True)
            log.debug("Reference added for %s: %s. ref. frame %s ",
                      rb_id, rb_rec.get_name(), rb_rec)
            Tis = io.read_transforms(d[2])
            relative_movers[i].add_internal_transformations(rb_rec, Tis)
        # add regular movers for the rigid bodies that are neither moved
        # anchored nor moved relative to others
        regular_movers = []
        for is_anchored, rb in zip(self.anchored, self.rbs):
            if(not is_anchored):
                name = rb.get_name()
                if(not name in relative_names):
                    rb.set_coordinates_are_optimized(True)
                    log.debug("adding a RigidBodyMover for %s", name)
                    mv = core.RigidBodyMover(rb, max_translation, max_rotation)
                    regular_movers.append(mv)
        self.movers = regular_movers
        self.movers += relative_movers

    def set_moving_parameters(self, max_translations, max_rotations):
        self.max_translations = max_translations
        self.max_rotations = max_rotations
        log.debug("Maximum translations %s", self.max_translations)
        log.debug("Maximum rotations %s", self.max_rotations)

    def set_scoring_function(self, scoring_function):
        """Set the scoring function to be used by MonteCarlo."""
        self._scoring_function = scoring_function

    def run_monte_carlo(self):
        """
            Run MonteCarlo sampling to generate possible states for DOMINO
        """
        t0 = time.time()
        log.info("Running MonteCarlo")
        mc = core.MonteCarlo(self.model)
        mc.set_scoring_function(self._scoring_function)
        mc.set_return_best(True)
        for i in range(self.cycles):
            log.info("Cycle: %s", i)
            for iters, T, tr, rot in zip(self.iterations, self.temperatures,
                                         self.max_translations, self.max_rotations):
                for rb in self.rbs:
                    log.debug("%s %s", rb.get_name(), rb.get_reference_frame())
                self.set_movers(tr, rot)
                mc.add_movers(self.movers)
                mc.set_kt(T)
                mc.optimize(iters)
                for rb in self.rbs:
                    log.debug("%s %s", rb.get_name(), rb.get_reference_frame())

                mc.clear_movers()
        log.info("MonteCarlo run finished. Time %s", time.time() - t0)

    def run_monte_carlo_with_relative_movers(self):
        """
            Run MonteCarlo sampling to generate possible states for DOMINO
        """
        t0 = time.time()
        log.info("Running MonteCarlo")
        mc = core.MonteCarlo(self.model)
        mc.set_scoring_function(self._scoring_function)
        mc.set_return_best(True)

        for i in range(self.cycles):
            log.info("Cycle: %s", i)
            for iters, T, tr, rot in zip(self.iterations, self.temperatures,
                                         self.max_translations, self.max_rotations):
                log.debug("BEFORE RELATIVE MOVERS")
                self.set_relative_movers(tr, rot)
                for m in self.movers:
                    m.propose()
                    m.accept()
                mc.add_movers(self.movers)
                mc.set_kt(T)
                log.debug("Optimizing for %s iterations ...", iters)
                mc.optimize(iters)
                log.debug("Finished optimizing.")
                mc.clear_movers()
        log.info("MonteCarlo run finished. Time %s", time.time() - t0)



## \example core/ms_connectivity_restraint.py
# This example shows how to use the MSConnectivityRestraint to ensure that all the particles that are part of complexes end up in a connected conformation following the optimization. It allows multiple copies of particles and takes an experimental tree as an input.
#

# -- File: ms_connectivity_restraint.py --#

import IMP
import IMP.core
import IMP.algebra
import sys

IMP.setup_from_argv(sys.argv, "ms connectivity restraint")

# Setup model

m = IMP.Model()
ps = [IMP.Particle(m) for x in range(6)]
ds = []
ds.append(IMP.core.XYZ.setup_particle(
    ps[0], IMP.algebra.Vector3D(0.0, 0.0, 0.0)))
ds.append(IMP.core.XYZ.setup_particle(
    ps[1], IMP.algebra.Vector3D(1.0, 1.0, 0.0)))
ds.append(IMP.core.XYZ.setup_particle(
    ps[2], IMP.algebra.Vector3D(2.0, 0.0, 0.0)))
ds.append(IMP.core.XYZ.setup_particle(
    ps[3], IMP.algebra.Vector3D(3.0, 0.0, 0.0)))
ds.append(IMP.core.XYZ.setup_particle(
    ps[4], IMP.algebra.Vector3D(4.0, -1.0, 0.0)))
ds.append(IMP.core.XYZ.setup_particle(
    ps[5], IMP.algebra.Vector3D(1000, 1000, 1000)))

# Create MS connectivity restraint

ub = IMP.core.HarmonicUpperBound(1.0, 0.1)
ss = IMP.core.DistancePairScore(ub)
r = IMP.core.MSConnectivityRestraint(m, ss)

# Add particle types to the restraint
# add_type() returns a unique type handle that can be used as an argument
# to add_composite() later on.

pa = r.add_type([ds[0], ds[1]])
pb = r.add_type([ds[2], ds[3]])
pc = r.add_type([ds[4]])
pd = r.add_type([ds[5]])

# Enter experimental tree data into restraint
# In add_composite(), the first argument is node label and the second
# argument is the parent.

i1 = r.add_composite([pa, pa, pb, pb, pc])
i2 = r.add_composite([pa, pb, pb, pc], i1)
i3 = r.add_composite([pa, pa, pb, pb], i1)
i4 = r.add_composite([pa, pb], i1)
i5 = r.add_composite([pa, pb, pb], i2)
i6 = r.add_composite([pb, pc], i2)
i7 = r.add_composite([pa, pa, pb], i3)
i8 = r.add_composite([pa, pb], i5)

# Evaluate the restraint score
r.evaluate(False)



#!/usr/bin/env python

""" Compute the dependencies of a swig file by running swig.
"""

import sys
from optparse import OptionParser
import os.path
import tools
import tools.thread_pool


parser = OptionParser()
parser.add_option("--build_dir", help="IMP build directory", default=None)
parser.add_option("--module_name", help="Module name", default=None)
parser.add_option("--include", help="Extra header include path", default=None)
parser.add_option("-s", "--swig",
                  dest="swig", default="swig",
                  help="The name of the swig command.")
parser.add_option("-b", "--build_system",
                  dest="build_system", help="The build system being used.")


def _fix(name, bs):
    if os.path.isabs(name):
        return name
    elif bs == "scons":
        return "#/build/" + name
    else:
        return os.path.join(os.getcwd(), "%s") % name


def get_dep_merged(finder, modules, name, extra_data_path):
    ret = []
    alldeps = finder.get_all_dependencies(modules)
    for d in alldeps:
        info = tools.get_dependency_info(d, extra_data_path)
        # cmake lists are semicolon-separated
        lst = tools.split(info[name], ';')
        ret.extend(lst)
    ret = sorted(set(ret))
    return ret


def setup_one(finder, module, build_system, swig, extra_data_path, include):
    includepath = get_dep_merged(finder, [module], "includepath",
                                 extra_data_path)
    swigpath = get_dep_merged(finder, [module], "swigpath", extra_data_path)

    with open("src/%s_swig.deps.in" % module.name, "w") as _:
        pass
    baseincludes = []
    if include:
        baseincludes.append(include)
    baseincludes.extend(('include', 'swig'))
    cmd = [swig, "-python", "-MM"] + ["-I" + x for x in baseincludes] \
        + ["-ignoremissing"] \
        + ["-I" + x for x in swigpath] + ["-I" + x for x in includepath] \
        + ["swig/IMP_%s.i" % module.name]

    lines = tools.run_subprocess(cmd).split("\n")
    names = []
    for x in lines:
        if x.endswith("\\"):
            x = x[:-1]
        x = x.strip()
        if not x.endswith(".h") and not x.endswith(".i") \
                and not x.endswith(".i-in"):
            continue
        names.append(x)

    final_names = [_fix(x, build_system) for x in names]
    final_list = "\n".join(final_names)
    tools.rewrite("src/%s_swig.deps" % module.name, final_list)


def main():
    (options, args) = parser.parse_args()
    mf = tools.ModulesFinder(configured_dir="build_info",
                             external_dir=options.build_dir,
                             module_name=options.module_name)
    pool = tools.thread_pool.ThreadPool()
    for m in [x for x in mf.values()
              if not isinstance(x, tools.ExternalModule) and x.ok
              and not x.python_only]:
        pool.add_task(setup_one, mf, m, options.build_system, options.swig,
                      options.build_dir, options.include)
    err = pool.wait_completion()
    if err:
        sys.stderr.write(err + '\n')
        return 1
    return 0


if __name__ == '__main__':
    main()



from __future__ import print_function
import IMP
import IMP.test
import IMP.core
import IMP.container


class SingletonTestModifier(IMP.SingletonModifier):

    def __init__(self, k):
        IMP.SingletonModifier.__init__(self)
        self.k = k

    def do_show(self, fh):
        fh.write("Test Particle")

    def apply_index(self, m, a0, da=None):
        if m.get_has_attribute(self.k, a0):
            pass
        else:
            m.add_attribute(self.k, a0, 1)

    def get_version_info(self):
        return 1

    def do_get_inputs(self, m, pis):
        return [m.get_particle(i) for i in pis]

    def do_get_outputs(self, m, pis):
        return [m.get_particle(i) for i in pis]


class PairTestModifier(IMP.PairModifier):

    def __init__(self, k):
        IMP.PairModifier.__init__(self)
        self.k = k
        self.sm = SingletonTestModifier(k)

    def do_show(self, fh):
        fh.write("Test Particle")

    def apply_index(self, m, a0, da=None):
        self.sm.apply_index(m, a0[0])
        self.sm.apply_index(m, a0[1])

    def get_version_info(self):
        return 1

    def do_get_inputs(self, m, pis):
        return [m.get_particle(i) for i in pis]

    def do_get_outputs(self, m, pis):
        return [m.get_particle(i) for i in pis]


def particle_has_attribute(p, k):
    return p.has_attribute(k)


def particle_pair_has_attribute(p, k):
    return p[0].has_attribute(k) and p[1].has_attribute(k)


def particle_triplet_has_attribute(p, k):
    return (
        p[0].has_attribute(k) and p[1].has_attribute(
            k) and p[2].has_attribute(k)
    )

# This file is generated by the make-container script


class Tests(IMP.test.TestCase):

    """Tests for ClassnameContainer related objects"""

    def create_particle(self, m):
        p = IMP.Particle(m)
        p.add_attribute(IMP.FloatKey("thekey"), float(1))
        return p

    def create_particle_pair(self, m):
        p0 = IMP.Particle(m)
        p1 = IMP.Particle(m)
        d0 = IMP.core.XYZ.setup_particle(p0)
        d1 = IMP.core.XYZ.setup_particle(p1)
        d0.set_coordinates(IMP.algebra.Vector3D(0, 0, 1))
        d1.set_coordinates(IMP.algebra.Vector3D(0, 0, 0))
        return (p0, p1)

    def same(self, a, b):
        return a.get_name() == b.get_name()

    def create_singleton_score_state(self, f, a, m, t):
        return IMP.core.SingletonConstraint(f, a, m, t.get_index())

    def create_pair_score_state(self, f, a, m, t):
        return IMP.core.PairConstraint(f, a, m, [x.get_index() for x in t])

    def create_particle_score(self):
        uf = IMP.core.Linear(0, 1)
        return IMP.core.AttributeSingletonScore(uf, IMP.FloatKey("thekey"))

    def create_particle_pair_score(self):
        uf = IMP.core.Linear(0, 1)
        return IMP.core.DistancePairScore(uf)

    def test_set(self):
        """Testing ClassnamesConstraint"""
        # write increment an int field
        # call evaluate and check that it is incremented
        IMP.set_log_level(IMP.VERBOSE)
        print("start")
        m = IMP.Model()
        print("hi")
        c = IMP.container.ListClassnameContainer(m)
        cs = []
        for i in range(0, 30):
            t = self.create_FUNCTIONNAME(m)
            if 'FUNCTIONNAME' == 'particle':
                c.add(t.get_index())
            else:
                c.add([x.get_index() for x in t])
            cs.append(t)
        print("dl")
        k = IMP.IntKey("thevalue")
        f = ClassnameTestModifier(k)  # noqa: F821
        print("apply")
        s = IMP.container.ClassnamesConstraint(f, None, c)
        print("add")
        m.add_score_state(s)
        m.update()
        for p in cs:
            self.assertTrue(FUNCTIONNAME_has_attribute(p, k))  # noqa: F821
        print("done")

    def test_sset(self):
        """Testing ClassnameConstraint"""
        # write increment an int field
        # call evaluate and check that it is incremented
        IMP.set_log_level(IMP.VERBOSE)
        print("start")
        m = IMP.Model()
        print("hi")
        t = self.create_FUNCTIONNAME(m)
        print("dl")
        k = IMP.IntKey("thevalue")
        f = ClassnameTestModifier(k)  # noqa: F821
        print("apply")
        s = self.create_CLASSFUNCTIONNAME_score_state(f, None, m, t)
        m.add_score_state(s)
        print("add")
        m.update()
        self.assertTrue(FUNCTIONNAME_has_attribute(t, k))  # noqa: F821
        print("done")


if __name__ == '__main__':
    IMP.test.main()



## \example kernel/flags.py
# A simple example of how to use IMP flags support in a python
# executable. Run with "--help" to see all available flags.

import IMP
import sys

IMP.add_string_flag("hello", "Hello world", "How to say hello")

IMP.setup_from_argv(sys.argv, "An example for using flags in python.")

if IMP.get_string_flag("hello") == "":
    IMP.write_help()
    exit(1)

print(IMP.get_string_flag("hello"))



## \example core/randomize_rigid_body.py
# This fragment shows how to either perturb or set the orientation of a rigid
# body randomly.

import IMP.core
import IMP.algebra
import IMP
import sys

IMP.setup_from_argv(sys.argv, "randomize rigid body")

m = IMP.Model()
p = IMP.Particle(m)
rbd = IMP.core.RigidBody.setup_particle(p, IMP.algebra.ReferenceFrame3D())
translation = IMP.algebra.get_random_vector_in(
    IMP.algebra.get_unit_bounding_box_3d())

# we don't yet have python code to generate a nearby rotation
rotation = IMP.algebra.get_random_rotation_3d()
transformation = IMP.algebra.Transformation3D(rotation, translation)
# Option 1:
# note, this overwrites the existing position
rbd.set_reference_frame(IMP.algebra.ReferenceFrame3D(transformation))
# Option 2:
# perturb the existing transformation
composed_tr = IMP.algebra.compose \
    (rbd.get_reference_frame().get_transformation_to(),
     transformation)
rbd.set_reference_frame(IMP.algebra.ReferenceFrame3D(composed_tr))
# Alternative to Option 2:
IMP.core.transform(rbd, transformation)
